
/* 
 Sivan Toledo, October 2000.

 Driver for the ILU factorization code based on the complete
 LU factorization code GP. 
 
 The code reads a matrix A from a file with (i,j,value) triplet in each
 line, 1-based indexing.
 The code generates a random vector x and computes b=A*x;
 The code optionally computes a column ordering for A using COLAMD.
 The code factors the matrix with drop tolerance.
 If the drop tolerance is 0.0 (no dropping), the factorization is a complete
 factorization and the code solves for y=A\b.
 If the drop tolerance is > 0, the code calls a bicgstab iterative solver
 with the incomplete factorization as a preconditioner.

 COLAMD if from:
 http://www.netlib.org/linalg/colamd/
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <assert.h>
#include "colamd.h"

#define DESC_TYPE_CSC 10001
#define CSC_DESC_SIZE 6

#define CSC_DESC_TYPE 0
#define CSC_DESC_M    1
#define CSC_DESC_N    2
#define CSC_DESC_NNZ  3
#define CSC_DESC_BASE 4
#define CSC_DESC_COLPTR 5
#define CSC_DESC_ROWIND(desc) (5+((desc)[2])+1)

static void statistics_reporter(double ctxt[], char* name, double* v)
{
  if (! strcmp(name,"FLOPS") )    ctxt[0] = *v;
  if (! strcmp(name,"NONZEROS") ) ctxt[1] = *v;
}

#define max(x,y) ((x)>(y) ? (x) : (y))
#define min(x,y) ((x)<(y) ? (x) : (y))

/*  csc_axpy(n, n, nz, a_desc, p_hat, v);*/
void csc_axpy(int m, int n,
	      double nz[],
	      int a_desc[],
	      double x[],    /* factor */
	      double b[])    /* output vector, b+=A*x */
{
  int i,j,p;
  double v;

  /*printf("base = %d\n",a_desc[CSC_DESC_BASE]);*/

  for (j=0; j<n; j++) {
    for (p = a_desc[CSC_DESC_COLPTR + j]   - a_desc[CSC_DESC_BASE] ; 
	 p < a_desc[CSC_DESC_COLPTR + j+1] - a_desc[CSC_DESC_BASE] ; 
	 p++) {
      i = a_desc[CSC_DESC_ROWIND(a_desc) + p] - a_desc[CSC_DESC_BASE];
      v = nz[p];
      /*
      if (a_desc[CSC_DESC_BASE] == 1) {
	printf("i j v p = %d %d %.2le %d\n",i,j,v,p);
      }
      */
	
      b[i] += v*x[j];
    }
  }
}

double dmax(int n, double v[])
{
  double m = -HUGE_VAL; /* -infinity */
  int i;
  for (i=0; i<n; i++) m = max(v[i],m);
  return m;
}

extern double dnrm2_();


double xdnrm2_(int* N, double v[], int* stride) /* inaccurate! use the blas */
{
  int n = *N;
  double m = 0.0;
  int i;
  for (i=0; i<n; i++) m += v[i] * v[i];
  m = sqrt(m);
  /*printf("dnrm2 = %lg\n",m);*/
  return m;
}
  
void* gp;

bicgstab(int n, double nz[], int a_desc[], 
	 void* lu, 
	 double b[], 
	 double x[],
	 double tol, int max_iters)
{
  int ione = 1;
  int i;
  double resid, bnrm, dtmp, omega, rho, rho_1, alpha, beta, error;
  int iter, flag;
  int ierr;
  double* r;
  double* r_tld;
  double* p;
  double* p_hat;
  double* v;
  double* s;
  double* s_hat;
  double* t;
  double flops_nnz[2];

  printf("allocating n=%d\n",n);
  r     = (double*) malloc(n*sizeof(double));
  r_tld = (double*) malloc(n*sizeof(double));
  p     = (double*) malloc(n*sizeof(double));
  p_hat = (double*) malloc(n*sizeof(double));
  v     = (double*) malloc(n*sizeof(double));
  s     = (double*) malloc(n*sizeof(double));
  s_hat = (double*) malloc(n*sizeof(double));
  t     = (double*) malloc(n*sizeof(double));
  printf("done allocating v=%08x\n",v);

  if (!r || !r_tld || !p || !p_hat || !v || !s || !s_hat || !t) {
    printf("could not allocate memory in bicgstab\n");
    exit(1);
  }

  iter = 0;
  flag = 0;
  
  bnrm = dnrm2_(&n,b,&ione);
  if (bnrm == 0.0) bnrm = 1.0;

  /* r = b - A*x */
  for (i=0; i<n; i++) r[i] = -b[i];
  csc_axpy(n, n, nz, a_desc, x, r);
  for (i=0; i<n; i++) r[i] = -r[i];

  error = dnrm2_(&n,r,&ione);
  if (error < tol) {
    printf("exact initial guess\n");
    free(r);
    free(r_tld);
    free(p);  
    free(p_hat);
    free(v);  
    free(s);  
    free(s_hat);
    free(t);  
    return;
  }
  /* omega = 1.0;*/

  /* r_tld = r;*/
  for (i=0; i<n; i++) r_tld[i] = r[i];

  /*
  printf("creating gp\n");
  gp_create_c(&gp,&ierr);
  gp_set_statistics_reporter_c(gp,&statistics_reporter,flops_nnz);
  printf("done creating gp\n");
  */

  for (iter=0; iter<max_iters; iter++) {
    /* rho   = ( r_tld'*r );*/
    for (rho=0.0, i=0; i<n; i++) rho += r_tld[i] * r[i];
    if ( rho == 0.0 ) break;

    if ( iter > 0 ) {
      beta  = ( rho/rho_1 )*( alpha/omega );
      /* p = r + beta*( p - omega*v );*/
      for (i=0; i<n; i++) p[i] = r[i] + beta*( p[i] - omega*v[i] );
    } else {
      /* p = r; */
      for (i=0; i<n; i++) p[i] = r[i];
    }
 
    /* p_hat = M \ p;*/
    for (i=0; i<n; i++) p_hat[i] = p[i];

    dgstrs_gp_c(gp,
		"No Transpose",
		n,1,
		lu,1,1,
		p_hat,1,1,NULL,
		&ierr);

    /* v = A*p_hat; */
    for (i=0; i<n; i++) v[i] = 0.0;
    csc_axpy(n, n, nz, a_desc, p_hat, v);

    /* alpha = rho / ( r_tld'*v ); */
    for (dtmp=0.0, i=0; i<n; i++) dtmp += r_tld[i] * v[i];
    alpha = rho / dtmp;
    /*printf("alpha >> %lg\n",alpha);*/

    /* s = r - alpha*v; */
    for (i=0; i<n; i++) s[i] = r[i] - alpha*v[i];

    if ( (error = dnrm2_(&n,s,&ione)/bnrm) < tol ) { /* early convergence check */
      /*  x = x + alpha*p_hat; */
      for (i=0; i<n; i++) x[i] += alpha*p_hat[i];
      for (i=0; i<n; i++) r[i] = s[i];               /* not really necessary */
      break;
    }
    /*printf("%04d (1): relative residual %lg\n",iter,error);*/

    /* s_hat = M \ s;                                 % stabilizer */
    for (i=0; i<n; i++) s_hat[i] = s[i];
    dgstrs_gp_c(gp,
		"No Transpose",
		n,1,
		lu,1,1,
		s_hat,1,1,NULL,
		&ierr);

    /* t = A*s_hat; */
    for (i=0; i<n; i++) t[i] = 0.0;
    csc_axpy(n, n, nz, a_desc, s_hat, t);

    /* omega = ( t'*s) / ( t'*t ); */
    for (dtmp =0.0, i=0; i<n; i++) dtmp  += t[i] * t[i];
    for (omega=0.0, i=0; i<n; i++) omega += t[i] * s[i];
    omega = omega / dtmp;

    /* x = x + alpha*p_hat + omega*s_hat;             % update approximation */
    /* r = s - omega*t;*/
    /*printf("alpha = %lg omega = %lg\n",alpha,omega);*/
    for (i=0; i<n; i++) x[i] += alpha*p_hat[i] + omega*s_hat[i];
    /*for (i=0; i<n; i++) r[i] = s[i] - omega*t[i];*/
    for (i=0; i<n; i++) r[i] -= alpha*v[i] + omega*t[i];

    error = dnrm2_(&n, r, &ione) / bnrm; /* check convergence */
    /*printf("%04d (2): relative residual %lg (rnorm=%lg)\n",iter,error,dnrm2_(&n, r, &ione));*/
    if ( error <= tol ) {
      printf("convergence, breaking loop\n");
      break;
    }
    if ( omega == 0.0 ) break;
    rho_1 = rho;
  }

  /* compute exact residual */
  /*
  for (i=0; i<n; i++) p[i] = 0.0;
  csc_axpy(n,n,nz,a_desc,x,p);
  for (i=0; i<n; i++) t[i] = p[i] - b[i];
  printf("bnrm = %lg real rnorm = %lg, rnorm = %lg relative normwise backward error %lg\n",
	 bnrm,
	 dnrm2_(&n,t,&ione),
	 dnrm2_(&n,r,&ione),
	 dnrm2_(&n,r,&ione) / dnrm2_(&n,b,&ione));
  */

  if ( error <= tol ) { /* converged */
    printf("converged, residual=%lg iters=%d\n",error,iter);
    flag = 0;
  } else if ( omega == 0.0 ) { /*  breakdown */
    printf("breakdown 2, iters=%d\n",iter);
    flag = -2;
  } else if ( rho == 0.0 ) {
    printf("breakdown 1, iters=%d\n",iter);
    flag = -1;
  } else { /* no convergence */
    printf("no convergence, residual=%lg iters=%d\n",error,iter);
    flag = 1;
  }

  free(r);
  free(r_tld);
  free(p);  
  free(p_hat);
  free(v);
  free(s);
  free(s_hat);
  free(t);
  printf("bicgstab done\n");

}

main(int argc, char* argv[])
{
  /*
  void*           gp;
  */

  void*   lu;
  int*    perm_c;
  int     perm_c_length;

  int*            p; /* column permutation */
  int             n,m,nnz;
  int             i,j,k;
  double          v;
  double*         nz;
  int*            a_desc;
  int             desc_ptr;
  char*           fname = NULL;
  FILE*           f;
  int*            clenptr; /* column lengths and pointers */
  double flops_nnz[2];
  double time;
  int ierr;

  double* x;
  double* y;
  double* b;
  double rerr,max_rerr;

  double dt = 0.0;

  int ione = 1;
  int use_colamd = 0;

  for (i=1; i<argc; i++) {
    if (!strcmp(argv[i],"-inputfile")) {
      i++;
      fname = argv[i];
    }
    if (!strcmp(argv[i],"-dt")) {
      i++;
      sscanf(argv[i],"%lg",&dt);
    }
    if (!strcmp(argv[i],"-colamd")) use_colamd = 1;
  }

  if ( !fname ) {
    printf("usage: %s -inputfile <filename> [-dt <droptolerance>] [-colamd]\n",argv[0]);
    printf("       <droptolerance> in [0.0:1.0]; 0 means don't drop, 1 is ILU(0)\n");
    exit(1);
  }

  printf("trying to open <%s>\n",fname);
  if ((f = fopen(fname,"r")) == NULL) {
    printf("input file not found\n");
    exit(1);
  }

  n = m = nnz = 0;
  while (!feof(f)) {
    if (fscanf(f,"%d %d %lf\n",&i,&j,&v) != 3) break;
    n = max(n,j);
    m = max(m,i);
    nnz++;
  }

  printf("rows=%d cols=%d nnz=%d\n",m,n,nnz);

  x = (double*) malloc( m * sizeof(double) ); 
  y = (double*) malloc( m * sizeof(double) ); 
  b = (double*) malloc( m * sizeof(double) ); 

  srandom(18733);
  for (i=0; i<m; i++) x[i] = (double) random() / (double) RAND_MAX;
  for (i=0; i<m; i++) y[i] = b[i] = 0.0;

  clenptr = (int*) malloc( (n + 1)                       * sizeof(int) ); 
  a_desc  = (int*) malloc( (CSC_DESC_SIZE + n + 1 + nnz) * sizeof(int) ); 
  nz      = (double*) malloc( nnz * sizeof( double ) ); 

  perm_c_length = m; 
  perm_c = (int*) malloc(m * sizeof(int));

  a_desc[CSC_DESC_TYPE] = DESC_TYPE_CSC;
  a_desc[CSC_DESC_M]    = m;
  a_desc[CSC_DESC_N]    = n;
  a_desc[CSC_DESC_NNZ]  = nnz;
  a_desc[CSC_DESC_BASE] = 0;

  desc_ptr = CSC_DESC_ROWIND(a_desc);

  /* now measure the size of each column */

  rewind(f);
  for (j=0; j<n; j++) clenptr[j] = 0; 
  for (k=0; k<nnz; k++) {
    if (fscanf(f,"%d %d %lf\n",&i,&j,&v) != 3) {
      printf("input file trancated\n");
      exit(1);
    }
    i--; j--; /* make them 0-based */
    ( clenptr[j] )++;
  }
  /* just check */
  k = 0;
  for (j=0; j<n; j++) 
    k += clenptr[j];
  assert(k == nnz);

  /* now compute column pointers */

  k = 0;
  for (j=0; j<n; j++) {
    int tmp;
    tmp =  clenptr[j];
    clenptr[j] = a_desc[CSC_DESC_COLPTR+j] = k;
    k += tmp;
  }
  clenptr[n] = a_desc[CSC_DESC_COLPTR+n] = k;
  assert(clenptr[n] == nnz);

  /* now read matrix into data structure */
  rewind(f);
  for (k=0; k<nnz; k++) {
    if (fscanf(f,"%d %d %lf\n",&i,&j,&v) != 3) {
      printf("input file trancated\n");
      exit(1);
    }
    i--; j--; /* make them 0-based */
    assert(i < n);
    assert(j < n);
    nz[ clenptr[j] ] = v;
    a_desc[ CSC_DESC_ROWIND(a_desc) + clenptr[j] ] = i;
    clenptr[j] ++;
  }

  /* compute b = A*x */

  csc_axpy(m,
	   n,
	   nz,
	   a_desc,
	   x,
	   b);
  printf("nrmb=%lg nrmx=%lg\n",dnrm2_(&n,b,&ione),dnrm2_(&n,x,&ione));

  /* now determine the column ordering */

  p = (int*) malloc((n+1) * sizeof(int));

  if (use_colamd) {
    double knobs[COLAMD_KNOBS];
    int    Alen;
    int*   A;

    Alen = colamd_recommended(nnz, m, n);
    A = malloc(Alen * sizeof(int)); assert(A);
    colamd_set_defaults (knobs) ;

    for (i=0; i<=n; i++)  p[i] = a_desc[ CSC_DESC_COLPTR + i ];
    for (k=0; k<nnz; k++) A[k] = a_desc[ CSC_DESC_ROWIND(a_desc) + k];

    if (!colamd (m, n, Alen, A, p, knobs)) {
      printf("colamd failed\n");
      exit(1);
    }
  } else {
    for (i=0; i<n; i++) p[i] = i; /* identity column permutation; should use colamd etc */
  }
    
  for (i=0; i<n; i++) perm_c[p[i]] = i;
  for (i=0; i<n; i++) perm_c[i] = p[i];

  printf("gp: factor\n");

  gp_create_c(&gp,&ierr); 
  gp_set_pivot_policy_c(gp,1);
  gp_set_pivot_threshold_c(gp,1.0); /* normal partial pivoting */
  gp_set_statistics_reporter_c(gp,&statistics_reporter,flops_nnz);
  gp_set_col_perm_c(gp,perm_c,perm_c_length,0,ierr);
  gp_set_drop_threshold_c(gp,dt); /* 0.0 means no dropping; 1.0 is ILU(0) */
  gp_set_col_fill_ratio_c(gp,-1);  /* don't do fill-ratio dropping */

  printf("factor\n");
  dgstrf_gp_c(gp,     /* factorization object */
	      m,      /* number of rows       */
	      n,      /* number of columns    */
	      nz,     /* numerical values     */
	      a_desc, /* descriptor           */
	      &lu,    /* dgstrf will set this pointer to an opaque object */
	      &ierr); /* error                */
  printf("factorization flop count %.0lfM, nnz = %.0lf\n",
	 flops_nnz[0]/1e6, flops_nnz[1]);

  /*gp_destroy_c(gp);*/

  if (dt == 0.0) {
    printf("dt==0.0 ==> nothing dropped => simple triangular solve\n");

    /*
    gp_create_c(&gp,&ierr);
    gp_set_statistics_reporter_c(gp,&statistics_reporter,flops_nnz);
    */
    for (i=0; i<n; i++) y[i] = b[i];
    dgstrs_gp_c(gp,
		"No Transpose",
		m,1,
		lu,1,1,
		y,1,1,NULL,
		&ierr);
    gp_destroy_c(gp);
  } else {
    printf("dt>0.0 calling bicgstab, initial guess=0\n");
    for (i=0; i<n; i++) y[i] = 0.0;
    bicgstab(n, nz, a_desc, 
	     lu, 
	     b, 
	     y, /* output */
	     1e-8, /* tolerance */
	     250); /* max iters */
  }

  max_rerr = 0.0;
  for (i=0; i<m; i++) {
    rerr = fabs( (x[i] - y[i]) / x[i] );
    max_rerr = max(rerr,max_rerr);
  }
  printf("max relative forward error is %lg\n",max_rerr);

  /* x=A*y; */
  for (i=0; i<n; i++) x[i] = 0.0;
  csc_axpy(m,n,nz,a_desc,y,x);
  max_rerr = 0.0;
  for (i=0; i<m; i++) {
    rerr = fabs( (x[i] - b[i]) / b[i] );
    max_rerr = max(rerr,max_rerr);
  }
  printf("max relative backward error is %lg\n",max_rerr);

  for (i=0; i<n; i++) y[i] = x[i] - b[i];
  printf("max normwise backward error is %lg\n",
	 dnrm2_(&n,y,&ione) / dnrm2_(&n,b,&ione));

  gstff_gp_c(lu);
}










